# coding: utf-8
# vi:set ts=3 sw=3:
# vim:set sts=0 noet:

=begin
Copyright (c) 2009 KISHIMOTO, Makoto

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
=end

require 'narray'

# downhill simplex method
# porting of http://aoki2.si.gunma-u.ac.jp/R/simplex.html
MAXIT=10000
EPSILON=1e-7
LO=0.8
HI=1.2
# start = initial vector
# residual = evaluation function. see sample.
def dh_simplex start, &residual
	ip = start.length
	ip1 = ip + 1
	ip2 = ip1 + 1
	ip3 = ip2 + 1
	columns = [start]
	(1...ip3).each do
		columns << start.dup
	end
	pa = NMatrix[*columns].transpose
	(0...pa.shape[1]).each do|i|
		pa[i, i] *= LO + rand * (HI - LO)
	end
	res = []
	(0...ip1).each do|c|
		res << residual.call(pa[c, true].to_a.flatten)
	end
	res.concat [0.0, 0.0]
	res = NArray[*res]
	mx = nil
	mi = nil
	(0...MAXIT).each do
		res0 = res[0...ip1]
		mx = 0
		(0...res0.shape[0]).each do|i|
			if res0[i] > res0[mx] then
				mx = i
			end
		end
		mi = 0
		(0...res0.shape[0]).each do|i|
			if res0[i] < res0[mi] then
				mi = i
			end
		end
		s = NArray[*pa[0...ip1, true].sum(0).to_a.flatten]
		if res[mx] < EPSILON or res[mi] < EPSILON or (res[mx]-res[mi])/res[mi] < EPSILON then
			break
		end
		i = ip2
		pa[ip2 - 1, true] = (2.0*s-ip2*NArray[*pa[mx, true].to_a.flatten])/ip
		res[ip2 - 1] = residual.call(pa[ip2 - 1, true].to_a.flatten)
		if res[ip2 - 1] < res[mi] then
			pa[ip3 - 1, true] = (3.0*s-(2.0*ip1+1.0)*NArray[*pa[mx, true].to_a.flatten])/ip
			res[ip3 - 1] = residual.call(pa[ip3 - 1, true].to_a.flatten)
			if res[ip3 - 1] <= res[ip2 - 1] then
				i = ip3
			end
		elsif res[ip2 - 1] > res[mx] then
			pa[ip3 - 1, true] = s/ip1
			res[ip3 - 1] = residual.call(pa[ip3 - 1, true].to_a.flatten)
			if res[ip3 - 1] >= res[mx] then
				(0...ip1).each do|index|
					if index == mi then
						next
					end
					pa[index, true] = (NArray[*pa[index, true].to_a.flatten]+NArray[*ps[mi, true]])*0.5
					res[index] = residual.call(pa[index, true].to_a.flatten)
				end
				i = 0 # false
			else
				i = ip3
			end
		end
		if i > 0 then
			pa[mx, true] = NArray[*pa[i - 1, true].to_a.flatten]
			res[mx] = res[i - 1]
		end
	end
	params = pa[mi, true].to_a.flatten
	residuals = residual.call params
	[params, residuals]
end

# sample
if $0 == __FILE__ then
	result = dh_simplex [80.0, 70.0, 0.5] do|p1, p2, p3|
		yy = []
		(1..10).each do|x|
			yy << p1/(1.0+p2*Math::E**(-p3*x))
		end
		res = 0.0
		[3.0, 8.0, 15.0, 35.0, 57.0, 80.0, 92.0, 95.0, 99.0, 100.0].zip yy do|y0, y|
			res += (y0-y)*(y0-y)
		end
		res
	end
	p result
end
